/*
	HID dumper
	Copyright 2000 (C) ITO, Takayuki
	All rights reserved

	platform independent
*/

#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include "hidparse.h"
#include "usage.h"

typedef void (*item_handler)(unsigned long);

unsigned short usage_page = 0;

const char *iof_name [][2] =
{
	{ "Data", "Constant" },
	{ "Array", "Variable" },
	{ "Absolute", "Relative" },
	{ "No Wrap", "Wrap" },
	{ "Linear", "Non Linear" },
	{ "Preferred State", "No Preferred" },
	{ "No Null position", "Null state" },
	{ "Non Volatile", "Volatile" },		/* Output/Feature only */
	{ "Bit Field", "Buffered Bytes" },
	{ NULL, NULL },
};

const char *collection_name [] =
{
	"Physical",
	"Application",
	"Logical",
	NULL
};

void handler_input (unsigned long value)
{
	int i, first = 1;
	unsigned long m;
	
	putchar ('(');
	if (value == 0x01)
		printf (iof_name [0][1]);
	else
	{
		for (i = 0, m = 1; i <= 8; i++, m <<= 1)
		{
			if (i <= 2 || (value & m))
			{
				if (!first)
					putchar (',');
				printf (iof_name [i][(value & m) > 0]);
				first = 0;
			}
		}
	}
	putchar (')');
}

void handler_output (unsigned long value)
{
	handler_input (value);
}

void handler_collection (unsigned long value)
{
	if (value < 3)
		printf (collection_name [value]);
}

void handler_feature (unsigned long value)
{
	handler_input (value);
}

void handler_end_collection (unsigned long value)
{
	/* none */
}

const char *main_item_tag_name [16] =
{
	NULL, NULL, NULL, NULL, 
	NULL, NULL, NULL, NULL, 
	"Input", 					/* important */
	"Output",  					/* important */
	"Collection",  				/* important */
	"Feature", 
	"End Collection", 			/* important */
	NULL, NULL, NULL
};

item_handler main_item_handler [16] =
{
	NULL, NULL, NULL, NULL,
	NULL, NULL, NULL, NULL,
	handler_input, 
	handler_output, 
	handler_collection, 
	handler_feature,
	handler_end_collection, 
	NULL, NULL, NULL,
};

void handler_usage_page (unsigned long value)
{
	const char *name = "(reserved)";
	
	usage_page = value;
	if (value <= 0x14)
		name = usage_page_name_0000 [value];
	else if (0x80 <= value && value <= 0x91)
		name = usage_page_name_0080 [value - 0x80];
	else if (value >= 0xff00)
		name = "vendor-defined";
	printf (name);
}

void handler_logical_minimum (unsigned long value)
{
	/* FIXME */
	printf ("%lu", value);
}

void handler_logical_maximum (unsigned long value)
{
	printf ("%lu", value);
}

void handler_physical_minimum (unsigned long value)
{
	printf ("%lu", value);
}

void handler_physical_maximum (unsigned long value)
{
	printf ("%lu", value);
}

void handler_unit_exponent (unsigned long value)
{
	/* FIXME */
	printf ("%lu", value);
}

void handler_unit (unsigned long value)
{
	/* FIXME */
	printf ("%lu", value);
}

void handler_report_size (unsigned long value)
{
	printf ("%lu", value);
}

void handler_report_id (unsigned long value)
{
	printf ("%lXh", value);
}

void handler_report_count (unsigned long value)
{
	printf ("%lu", value);
}

const char *global_item_tag_name [16] =
{
	"Usage Page", 				/* important */
	"Logical Minimum",  		/* important */
	"Logical Maximum",  		/* important */
	"Physical Minimum", 		/* important */
	"Physical Maximum",  		/* important */
	"Unit Exponent", 
	"Unit", 
	"Report Size", 				/* important */
	"Report ID",  				/* important */
	"Report Count", 			/* important */
	"Push", 
	"Pop", 
	NULL, NULL, NULL, NULL
};

item_handler global_item_handler [16] =
{
	handler_usage_page, 
	handler_logical_minimum, 
	handler_logical_maximum, 
	handler_physical_minimum,
	handler_physical_maximum, 
	handler_unit_exponent, 
	handler_unit, 
	handler_report_size,
	handler_report_id, 
	handler_report_count, 
	NULL /* handler_push */, 
	NULL /* handler_pop */,
	NULL, NULL, NULL, NULL
};

void handler_usage (unsigned long value)
{
	const char *name = NULL;
	
	/* FIXME extended usage */

	switch (usage_page)
	{
	case USAGE_PAGE_GENERIC_DESKTOP:
		if (value <= 0x08)
			name = generic_desktop_usage_name_0000 [value];
		else if (0x30 <= value && value <= 0x46)
			name = generic_desktop_usage_name_0030 [value - 0x30];
		else if (0x80 <= value && value <= 0x93)
			name = generic_desktop_usage_name_0080 [value - 0x80];
		if (name != NULL)
			printf (name);
		else
			printf ("(reserved)");
		break;
	case USAGE_PAGE_KEYBOARD:
		printf ("%lu", value);
		break;
	case USAGE_PAGE_LED:
		if (value <= 0x4d)
		{
			name = led_usage_name [value];
			if (name != NULL)
				printf (name);
			else
				printf ("(not implemented)");
		}
		else
			printf ("(reserved)");
		break;
	case USAGE_PAGE_BUTTONS:
		printf ("Button %lu", value);
		break;
	default:
		printf ("%lu", value);
		break;
	}
}

void handler_usage_minimum (unsigned long value)
{
	handler_usage (value);
}

void handler_usage_maximum (unsigned long value)
{
	handler_usage (value);
}

const char *local_item_tag_name [16] =
{
	"Usage", 					/* important */
	"Usage Minimum", 			/* important */ 
	"Usage Maximum", 			/* important */
	"Designator Index",
	"Designator Minimum", 
	"Designator Maximum", 
	NULL, 
	"String Index",
	"String Minimum", 
	"String Maximum", 
	"Delimiter", 
	NULL, NULL, NULL, NULL, NULL
};

item_handler local_item_handler [16] =
{
	handler_usage, 
	handler_usage_minimum, 
	handler_usage_maximum, 
	NULL /* handler_designator_index */,
	NULL /* handler_designator_minimum */, 
	NULL /* handler_designator_maximum */, 
	NULL, 
	NULL /* handler_string_index */,
	NULL /* handler_string_minimum */, 
	NULL /* handler_string_maximum */, 
	NULL /* handler_delimiter */, 
	NULL, NULL, NULL, NULL, NULL
};

int other_main(char *buf, size_t desc_len)
//int main (int argc, char **argv)
{
	int i, fd, first_report_id;
	//size_t desc_len, 
	
	size_t num_items, num_insns, total_report_size;
	decomp_item items [256];
	report_insn insns [256];
	//char buf [256];

	//fd = open (argv [1], O_RDONLY);
	//if (fd < 0)
	//	return 1;
	//desc_len = read (fd, buf, 256);
	//close (fd);

	num_items = 256;
	decompose_report_descriptor (buf, desc_len, items, &num_items);
	for (i = 0; i < num_items; i++)
	{
		int tag;
		decomp_item *item = &items [i];
		const char *name;
		item_handler handler;

		printf ("%3Xh(%3dd)  %02X ",
			item->offset, item->offset, item->prefix);
		switch (item->size)
		{
		case 0:
			printf ("           ");
			break;
		case 1:
			printf ("%02lX         ", item->value);
			break;
		case 2:
			printf ("%04lX       ", item->value);
			break;
		case 4:
			printf ("%08lX   ", item->value);
			break;
		}
		tag = item->tag;
		switch (item->type)
		{
		case ITEM_TYPE_MAIN:
			name = main_item_tag_name [tag];
			handler = main_item_handler [tag];
			break;
		case ITEM_TYPE_GLOBAL:
			name = global_item_tag_name [tag];
			handler = global_item_handler [tag];
			break;
		case ITEM_TYPE_LOCAL:
			name = local_item_tag_name [tag];
			handler = local_item_handler [tag];
			break;
		}
		printf ("%-20s", name);
		if (handler != NULL)
			handler (item->value);
		putchar ('\n');
	}

	num_insns = 256;
	parse_report_descriptor (items, num_items, insns, &num_insns, &total_report_size, &first_report_id);
	for (i = 0; i < num_insns; i++)
	{
		report_insn *insn = &insns [i];

		printf ("usage=%x:%x ",
			insn->usage_page,
			insn->usage_id);
		if (insn->is_phy_signed)
			printf ("min=%ld max=%ld ",
				insn->signed_phy_min,
				insn->signed_phy_max);
		else
			printf ("min=%lu max=%lu ",
				insn->phy_min,
				insn->phy_max);
		printf ("byte=%d bit=%d size=%x\n",
			insn->byte_idx,
			insn->bit_pos,
			insn->num_bits);
	}
	printf ("total %d bytes, report ID = %d\n", total_report_size, first_report_id);

	return 0;
}
