#include <USBKit.h>
#include <USB_hid.h>
#include <Joystick.h>
#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fstream>

#include "joystick_driver.h"

extern "C" {
#include "hidparse.h"
int other_main(char *buf, size_t desc_len);
};

using namespace std;

#if 0
typedef struct
{
	uint8	length;
	uint8	descriptor_type;
	uint16	hid_version;
	uint8	country_code;
	uint8	num_descriptors;
	struct
	{
		uint8	descriptor_type;
		uint16	descriptor_length;
	} _PACKED descriptor_info [1];
} _PACKED usb_hid_descriptor;

#define	USB_DESCRIPTOR_HID			0x21
#define	USB_DESCRIPTOR_HID_REPORT	0x22

//#define	USB_REQUEST_HID_GET_REPORT	0x01
#endif

#define DPRINTF_INFO printf

/* maximum number of axes on one controller (pads count as 2 axes each) */
#define MAX_AXES 12
/* maximum number of hats on one controller -- PADS SHOULD BE RETURNED AS AXES! */
#define MAX_HATS 8
/* maximum number of buttons on one controller */
#define MAX_BUTTONS 32
/* maximum number of controllers on one port */
#define MAX_STICKS 4

typedef struct my_device_info
{
	/* list structure */
	struct my_device_info *next;
	
	report_insn	*insns;
	size_t num_insns;
	size_t total_report_size;
	int num_buttons, num_axes, num_hats;
};

void do_driver(BUSBDevice *device, uint8 *rep_desc, size_t desc_len)
{
	size_t num_items, actual;
	decomp_item *items;
	report_insn *realloc_insns = 0;
	int report_id;
	
	my_device_info *my_dev = (my_device_info *)malloc(sizeof(my_device_info));
	
	/* decompose report descriptor */
	num_items = desc_len;	/* XXX */
	items = (decomp_item *)malloc (sizeof (decomp_item) * num_items);
	//assert (items != NULL);
	decompose_report_descriptor (rep_desc, desc_len, items, &num_items);
	//free (rep_desc);
	
	/* parse report descriptor */

	my_dev->num_insns = num_items;	/* XXX */
	my_dev->insns = (report_insn *)malloc (sizeof (report_insn) * my_dev->num_insns);
	//assert (my_dev->insns != NULL);
	
	parse_report_descriptor (items, num_items, my_dev->insns, 
		&my_dev->num_insns, &my_dev->total_report_size, &report_id);
	free (items);
	
	realloc_insns = (report_insn *)realloc (my_dev->insns, sizeof (report_insn) * my_dev->num_insns);
	my_dev->insns = realloc_insns;
	DPRINTF_INFO ("%d items, %d insns, %d bytes\n", 
		(int)num_items, (int)my_dev->num_insns, (int)my_dev->total_report_size);

	/* count axes, hats and buttons */

	count_controls (my_dev->insns, my_dev->num_insns,
		&my_dev->num_axes, &my_dev->num_hats, &my_dev->num_buttons);
	DPRINTF_INFO ("%d axes, %d hats, %d buttons\n",
		my_dev->num_axes, my_dev->num_hats, my_dev->num_buttons);
	if (my_dev->num_axes > MAX_AXES)
		my_dev->num_axes = MAX_AXES;
	if (my_dev->num_buttons > MAX_BUTTONS)
		my_dev->num_buttons = MAX_BUTTONS;
	if (my_dev->num_hats > MAX_HATS)
		my_dev->num_hats = MAX_HATS;
	
	/* get initial state */
	status_t st = device->ControlTransfer(
	USB_REQTYPE_INTERFACE_IN | USB_REQTYPE_CLASS,  // requestType
		B_USB_REQUEST_HID_GET_REPORT,  // request
		0x0100 | report_id, // value
		0, // index
		my_dev->total_report_size,  // length
		&actual); // data
		
	cout << "status initial result: " << st << endl;
		other_main((char *)rep_desc, desc_len);
}

void print(const usb_device_descriptor *d)
{
	cout << "begin : usb_device_descriptor" << endl;
	
	cout << hex << "length: " << (int)d->length << endl;
	cout << hex << "descriptor_type: " << (int)d->descriptor_type << endl;
	cout << hex << "usb_version: " << (int)d->usb_version << endl;
	cout << hex << "device_class: " << (int)d->device_class << endl;
	cout << hex << "device_subclass: " << (int)d->device_subclass << endl;
	cout << hex << "device_protocol: " << (int)d->device_protocol << endl;
	cout << hex << "max_packet_size_0: " << (int)d->max_packet_size_0 << endl;
	cout << hex << "vendor_id: " << (int)d->vendor_id << endl;
	cout << hex << "product_id: " << (int)d->product_id << endl;
	cout << hex << "device_version: " << (int)d->device_version << endl;
	cout << hex << "manufacturer: " << (int)d->manufacturer << endl;
	cout << hex << "product: " << (int)d->product << endl;
	cout << hex << "serial_number: " << (int)d->serial_number << endl;
	cout << hex << "num_configurations: " << (int)d->num_configurations << endl;
	
	cout << "end : usb_device_descriptor" << endl << endl;
}

void print(const usb_configuration_descriptor *c)
{
	cout << "begin : usb_configuration_descriptor" << endl;
	
	cout << hex << "length: " << (int)c->length << endl;
	cout << hex << "descriptor_type: " << (int)c->descriptor_type << endl;
	cout << hex << "total_length: " << (int)c->total_length << endl;
	cout << hex << "number_interfaces: " << (int)c->number_interfaces << endl;
	cout << hex << "configuration_value: " << (int)c->configuration_value << endl;
	cout << hex << "configuration: " << (int)c->configuration << endl;
	cout << hex << "attributes: " << (int)c->attributes << endl;
	cout << hex << "max_power: " << (int)c->max_power << endl;
	
	cout << "end : usb_configuration_descriptor" << endl << endl;
}


int main(int argc, char **argv)
{
	if(argc != 2) {
		cout << argv[0] << " <joystick/gamepad descriptor_file>" << endl;
		return 1;
	}
	ifstream fin(argv[1]);
	if(!fin) return 0;
	fin.seekg(0, ios::end);
	int len = fin.tellg();
	fin.seekg(0, ios::beg);
	
	uint8 buffer[len+1];
	fin.read((char *)buffer, len);
	fin.close();
	other_main((char *)buffer, len);
	return 0;
	
	usb_hid_descriptor hid_desc;
	size_t desc_len;
	uint8 *rep_desc = NULL;
	
	BUSBDevice device("/dev/bus/usb/4/0");
	
	cout.setf(ios::showbase);
	
	const usb_device_descriptor *dev_desc = device.Descriptor();
	
	const BUSBConfiguration *config = device.ConfigurationAt(0);
	const usb_configuration_descriptor *conf_desc = config->Descriptor();//device.ConfigurationAt(0)->Descriptor();
	
	print(dev_desc);
	print(conf_desc);
	
	cout << hex << "Vendor: " << dev_desc->vendor_id << endl;
	cout << device.CountConfigurations() << endl;
	
	//char 
	/* read HID descriptor */
	status_t st = device.ControlTransfer(
	USB_REQTYPE_INTERFACE_IN | USB_REQTYPE_STANDARD,  // requestType
		USB_REQUEST_GET_DESCRIPTOR,  // request
		B_USB_HID_DESCRIPTOR_HID << 8, // value
		0, // index
		sizeof(usb_hid_descriptor),  // length
		&hid_desc); // data
	
	cout << "Read HID Descriptor Result: " << st << endl;
	//cout << hex << hid_desc.descriptor_info [0].descriptor_length << endl;
	
	desc_len = hid_desc.descriptor_info [0].descriptor_length;
	rep_desc = new uint8[desc_len];
	
	/* read report descriptor */
	st = device.ControlTransfer(
	USB_REQTYPE_INTERFACE_IN | USB_REQTYPE_STANDARD,  // requestType
		USB_REQUEST_GET_DESCRIPTOR,  // request
		B_USB_HID_DESCRIPTOR_REPORT << 8, // value
		0, // index
		desc_len,  // length
		rep_desc); // data
	
#ifdef WRITE
	ofstream fout;
	fout.open("rep_desc.bin");
	
	fout.write(rep_desc, desc_len);
	fout.close();
#endif

	cout << "Read Report Descriptor Result: " << st << endl;
	
		//if(desc_le
		do_driver(&device, rep_desc, desc_len);
		
	cout << "num_descriptors: " << (int)desc_len << endl;
	//cout << config->CountInterfaces() << endl;
	
	
	//const BUSBInterface *interface = config->InterfaceAt(0);
	
	
	//cout << hex << (int)hid_desc.descriptor_type << endl;
	//const BUSBConfiguration *conf = device.ConfigurationAt(0);
	
	//const usb_configuration_descriptor *conf_desc = device.ConfigurationAt(0)->Descriptor();
	
	//cout << conf_desc->total_length << endl;
	//cout << hex << sizeof(usb_device_descriptor) << endl;
	//cout << "st: " << st << endl;
	//cout << sizeof(usb_hid_descriptor) << endl;
	//cout << device.Descriptor() << endl;
	
	if(rep_desc) delete [] rep_desc;
	
	//int test = -0x80;
	
	//cout << test << endl;
	
	BJoystick stick;
	char buf[256];
	cout << stick.CountDevices() << endl;
	stick.GetDeviceName(0, buf);
	
	cout << buf << endl;
	
	cout << sizeof(_joystick) << endl;
	return 0;
}
